@@ -12,7 +12,7 @@ use std::{ffi::CString, ops::Range};
 use fitsio::{errors::check_status as fits_check_status, FitsFile};
 use indicatif::ProgressStyle;
 use itertools::izip;
-use log::warn;
+use log::{trace, warn};
 use mwa_rust_core::{
     constants::VEL_C,
     erfa_sys,
@@ -84,6 +84,8 @@ fn decode_uvfits_baseline(bl: usize) -> (usize, usize) {
 
 /// A helper struct to write out a uvfits file.
 ///
+/// Note: only a single contiguous spectral window is supported.
+///
 /// TODO: make writer and reader a single class?
 pub struct UvfitsWriter<'a> {
     /// The path to the uvifts file.
@@ -96,7 +98,12 @@ pub struct UvfitsWriter<'a> {
     /// The number of uvfits rows that have currently been written.
     current_num_rows: usize,
 
-    /// The center frequency of the center fine channel \[Hz\].
+    /// The center frequency of the center fine channel of the spectral
+    /// window being written to this file. \[Hz\]
+    ///
+    /// This is used in both the reference frequency (`FREQ`) in the antenna HDU,
+    /// and the center reference value of the frequency axis (`CRVAL4`) in the
+    /// visibility hdu.
     centre_freq: f64,
 
     /// A `hifitime` [Epoch] struct associated with the first timestep of the
@@ -126,11 +133,12 @@ impl<'a> UvfitsWriter<'a> {
     /// let start_epoch = time::gps_to_epoch(first_gps_time);
     /// ```
     ///
-    /// `centre_freq_hz` is the centre frequency of the coarse band that this
-    /// uvfits file pertains to [Hz].
+    /// `centre_freq_hz` is center frequency of the center fine channel of the
+    /// spectral window being written to this file. \[Hz\]
     ///
-    /// `centre_freq_chan` is the index (from zero) of the centre frequency of
-    /// the channels of this uvfits file.
+    /// `centre_freq_chan` is the index (from zero) of the center frequency of
+    /// the center fine channel of the spectral] window being written to this
+    /// file.
     ///
     /// `phase_centre` is a [`RADec`] of the observation's phase center, used to
     /// populate the `OBSRA` and `OBSDEC` keys.
@@ -161,6 +169,7 @@ impl<'a> UvfitsWriter<'a> {
     ) -> Result<Self, UvfitsWriteError> {
         // Delete any file that already exists.
         if filename.exists() {
+            trace!("file {:?} exists, deleting", &filename);
             std::fs::remove_file(&filename)?;
         }
 
@@ -168,6 +177,7 @@ impl<'a> UvfitsWriter<'a> {
         let mut status = 0;
         let c_filename = CString::new(filename.to_str().unwrap())?;
         let mut fptr = std::ptr::null_mut();
+        trace!("initialising fits file with fitsio_sys ({:?})", &filename);
         unsafe {
             fitsio_sys::ffinit(
                 &mut fptr as *mut *mut _, /* O - FITS file pointer                   */
@@ -182,6 +192,7 @@ impl<'a> UvfitsWriter<'a> {
         let mut naxes = [0, 3, 4, num_chans as i64, 1, 1];
         let num_group_params = 5;
         let total_num_rows = num_timesteps * num_baselines;
+        trace!("setting group params in fits file ({:?})", &filename);
         unsafe {
             fitsio_sys::ffphpr(
                 fptr,                  /* I - FITS file pointer                        */
@@ -198,12 +209,14 @@ impl<'a> UvfitsWriter<'a> {
         fits_check_status(status)?;
 
         // Finally close the file.
+        trace!("closing fits file ({:?})", &filename);
         unsafe {
             fitsio_sys::ffclos(fptr, &mut status);
         }
         fits_check_status(status)?;
 
         // Open the fits file with rust-fitsio.
+        trace!("opening  ({:?})", &filename);
         let mut u = FitsFile::edit(&filename)?;
         let hdu = u.hdu(0)?;
         hdu.write_key(&mut u, "BSCALE", 1.0)?;
@@ -346,6 +359,13 @@ impl<'a> UvfitsWriter<'a> {
         let start_epoch = time::gps_to_epoch(first_gps_time);
         let phase_centre = RADec::from_mwalib_phase_or_pointing(&context.metafits_context);
         let centre_freq_chan = num_img_chans / 2;
+        // Get the first coarse chan index, and multiply by number of fine chans per coarse
+        // then add the centre_freq_chan which will give us the centre index (of all fine channels).
+        let img_centre_fine_chan_idx: usize =
+            mwalib_coarse_chan_range.start * fine_chans_per_coarse + centre_freq_chan;
+        let centre_freq_hz =
+            context.metafits_context.metafits_fine_chan_freqs_hz[img_centre_fine_chan_idx];
+
         Self::new(
             filename,
             mwalib_timestep_range.len(),
@@ -353,7 +373,7 @@ impl<'a> UvfitsWriter<'a> {
             num_img_chans,
             start_epoch,
             context.metafits_context.corr_fine_chan_width_hz as f64,
-            context.metafits_context.centre_freq_hz as f64,
+            centre_freq_hz,
             centre_freq_chan,
             phase_centre,
             Some(context.metafits_context.obs_name.as_str()),
@@ -967,6 +987,7 @@ mod tests {
 /// Tests which require the use of the aoflagger feature
 mod tests_aoflagger {
     use super::*;
+    use approx::assert_abs_diff_eq;
     use mwa_rust_core::{
         constants::{
             COTTER_MWA_HEIGHT_METRES, COTTER_MWA_LATITUDE_RADIANS, COTTER_MWA_LONGITUDE_RADIANS,
@@ -1013,11 +1034,7 @@ mod tests_aoflagger {
                     get_required_fits_key!($right_fptr, &$right_hdu, key),
                 ) {
                     (Ok::<String, _>(left_val), Ok::<String, _>(right_val)) => {
-                        assert_eq!(
-                            left_val, right_val,
-                            "mismatch for metafits short string key {}",
-                            key,
-                        );
+                        assert_eq!(left_val, right_val, "mismatch for short string key {}", key,);
                     }
                     (Err(err), Ok(right_val)) => {
                         panic!(
@@ -1043,7 +1060,7 @@ mod tests_aoflagger {
                     (Ok(left_val), Ok(right_val)) => {
                         assert!(
                             approx_eq!(f64, left_val, right_val, F64Margin::default()),
-                            "mismatch for metafits short f64 key {}. {} != {} (margin={:?})",
+                            "mismatch for short f64 key {}. {} != {} (margin={:?})",
                             key,
                             left_val,
                             right_val,
@@ -1238,9 +1255,11 @@ mod tests_aoflagger {
                 "BITPIX", "NAXIS", "NAXIS1", "NAXIS2", "NAXIS3", "NAXIS4", "NAXIS5", "NAXIS6",
                 "BSCALE", "PSCAL1", "PZERO1", "PSCAL2", "PZERO2", "PSCAL3", "PZERO3", "PSCAL4",
                 "PZERO4", "PSCAL5", "PZERO5", "CRVAL2", "CRPIX2", "CDELT2", "CRVAL3", "CDELT3",
-                "CRPIX3", "CRVAL4", "CDELT4", "CRPIX4", "CRVAL5", "CRPIX5", "CDELT5", "CRVAL6",
-                "CRPIX6", "CDELT6", "EPOCH", "OBSRA",
-                "OBSDEC",
+                "CRPIX3",
+                // This is actually incorrect in Cotter, see https://github.com/MWATelescope/Birli/issues/6
+                // "CRVAL4",
+                "CDELT4", "CRPIX4", "CRVAL5", "CRPIX5", "CDELT5", "CRVAL6", "CRPIX6", "CDELT6",
+                "EPOCH", "OBSRA", "OBSDEC",
                 // TODO: "FIBRFACT", (v-factor, cable type?)
             ],
             left_fptr,
@@ -1269,10 +1288,12 @@ mod tests_aoflagger {
 
         assert_f64_string_keys_eq!(
             vec![
-                "BITPIX", "NAXIS", "NAXIS1", "NAXIS2", "PCOUNT", "GCOUNT", "TFIELDS",
+                "BITPIX", "NAXIS", "NAXIS1", "NAXIS2", "PCOUNT", "GCOUNT", "TFIELDS", //
                 // TODO: "ARRAYX", "ARRAYY", "ARRAYZ",
-                "FREQ",   // TODO: "GSTIA0",
-                "DEGPDY", // TODO: "RDATE",
+                // "FREQ", // This is incorrect in Cotter. See: https://github.com/MWATelescope/Birli/issues/6
+                // TODO: "GSTIA0",
+                "DEGPDY", //
+                // TODO: "RDATE",
                 "POLARX", "POLARY", "UT1UTC", "DATUTC", "NUMORB", "NOPCAL", "FREQID", "IATUTC",
             ],
             left_fptr,
@@ -1317,7 +1338,7 @@ mod tests_aoflagger {
                 _ => {
                     panic!("could not read HDUs as group table")
                 }
-            };
+            }
         };
     }
 
@@ -1622,4 +1643,75 @@ mod tests_aoflagger {
         assert_uvfits_ant_header_eq(&mut birli_fptr, &mut cotter_fptr);
         assert_uvfits_ant_table_eq(&mut birli_fptr, &mut cotter_fptr);
     }
+
+    /// This test ensures center frequencies are calculated correctly.
+    /// See: https://github.com/MWATelescope/Birli/issues/6
+    #[test]
+    fn center_frequencies() {
+        let context = get_mwa_ord_context();
+
+        let tmp_uvfits_file = NamedTempFile::new().unwrap();
+
+        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        assert_eq!(img_timestep_idxs.len(), 4);
+        let img_timestep_range =
+            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
+        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let img_coarse_chan_range =
+            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let img_baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
+
+        let array_pos = Some(LatLngHeight {
+            longitude_rad: COTTER_MWA_LONGITUDE_RADIANS,
+            latitude_rad: COTTER_MWA_LATITUDE_RADIANS,
+            height_metres: COTTER_MWA_HEIGHT_METRES,
+        });
+
+        let mut u = UvfitsWriter::from_mwalib(
+            tmp_uvfits_file.path(),
+            &context,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            &img_baseline_idxs,
+            array_pos,
+        )
+        .unwrap();
+
+        let flag_array = init_flag_array(
+            &context,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            Some(get_antenna_flags(&context)),
+        );
+
+        let (jones_array, flag_array) = context_to_jones_array(
+            &context,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            Some(flag_array),
+        );
+
+        u.write_jones_flags(
+            &context,
+            &jones_array,
+            &flag_array,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            &img_baseline_idxs,
+        )
+        .unwrap();
+
+        u.write_ants_from_mwalib(&context.metafits_context).unwrap();
+
+        let mut birli_fptr = fits_open!(&tmp_uvfits_file.path()).unwrap();
+
+        let birli_vis_hdu = fits_open_hdu!(&mut birli_fptr, 0).unwrap();
+        let birli_vis_freq: f64 =
+            get_required_fits_key!(&mut birli_fptr, &birli_vis_hdu, "CRVAL4").unwrap();
+        assert_abs_diff_eq!(birli_vis_freq, 229760000.);
+        let birli_ant_hdu = fits_open_hdu!(&mut birli_fptr, 1).unwrap();
+        let birli_ant_freq: f64 =
+            get_required_fits_key!(&mut birli_fptr, &birli_ant_hdu, "FREQ").unwrap();
+        assert_abs_diff_eq!(birli_ant_freq, 229760000.);
+    }
 }
